@@ -3,7 +3,7 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE-BSD-3-Clause file.
 
-use std::io::{self, IoSlice, Write};
+use std::io::{self, IoSlice, Read, Write};
 use std::mem::size_of;
 use std::sync::Arc;
 use std::time::Duration;
@@ -11,15 +11,17 @@ use std::time::Duration;
 use vm_memory::ByteValued;
 
 use super::{
-    MetricsHook, Server, ServerUtil, ServerVersion, SrvContext, ZcReader, ZcWriter, DIRENT_PADDING,
-    MAX_BUFFER_SIZE, MAX_REQ_PAGES,
+    MetricsHook, Server, ServerUtil, ServerVersion, SrvContext, ZcReader, ZcWriter,
+    BUFFER_HEADER_SIZE, DIRENT_PADDING, MAX_BUFFER_SIZE, MAX_REQ_PAGES, MIN_READ_BUFFER,
 };
 use crate::abi::linux_abi::*;
 #[cfg(feature = "virtiofs")]
 use crate::abi::virtio_fs::{RemovemappingIn, RemovemappingOne, SetupmappingIn};
-use crate::api::filesystem::{DirEntry, Entry, FileSystem, GetxattrReply, ListxattrReply};
+use crate::api::filesystem::{
+    DirEntry, Entry, FileSystem, GetxattrReply, IoctlData, ListxattrReply,
+};
 use crate::async_util::AsyncDrive;
-use crate::transport::{FsCacheReqHandler, Reader, Writer};
+use crate::transport::{pagesize, FsCacheReqHandler, Reader, Writer};
 use crate::{bytes_to_cstr, encode_io_error_kind, BitmapSlice, Error, Result};
 
 impl<F: FileSystem + Sync, D: AsyncDrive> Server<F, D> {
@@ -623,39 +625,31 @@ impl<F: FileSystem + Sync, D: AsyncDrive> Server<F, D> {
             return ctx.reply_ok(Some(out), None);
         }
 
-        // These fuse features are supported by this server by default.
-        let supported = FsOptions::ASYNC_READ
-            | FsOptions::PARALLEL_DIROPS
-            | FsOptions::BIG_WRITES
-            | FsOptions::AUTO_INVAL_DATA
-            | FsOptions::ASYNC_DIO
-            | FsOptions::HAS_IOCTL_DIR
-            | FsOptions::MAX_PAGES
-            | FsOptions::EXPLICIT_INVAL_DATA
-            | FsOptions::PERFILE_DAX;
-
         let capable = FsOptions::from_bits_truncate(flags);
 
         match self.fs.init(capable) {
             Ok(want) => {
-                let enabled = capable & (want | supported);
+                let enabled = capable & want;
                 info!(
                     "FUSE INIT major {} minor {}\n in_opts: {:?}\nout_opts: {:?}",
                     major, minor, capable, enabled
                 );
 
-                let out = InitOut {
+                let mut out = InitOut {
                     major: KERNEL_VERSION,
                     minor: KERNEL_MINOR_VERSION,
                     max_readahead,
                     flags: enabled.bits(),
                     max_background: ::std::u16::MAX,
                     congestion_threshold: (::std::u16::MAX / 4) * 3,
-                    max_write: MAX_BUFFER_SIZE,
-                    time_gran: 1,             // nanoseconds
-                    max_pages: MAX_REQ_PAGES, // 1MB
+                    max_write: MIN_READ_BUFFER - BUFFER_HEADER_SIZE,
+                    time_gran: 1, // nanoseconds
                     ..Default::default()
                 };
+                if enabled.contains(FsOptions::MAX_PAGES) {
+                    out.max_pages = MAX_REQ_PAGES;
+                    out.max_write = MAX_REQ_PAGES as u32 * pagesize() as u32; // 1MB
+                }
                 let vers = ServerVersion { major, minor };
                 self.vers.store(Arc::new(vers));
                 if minor < KERNEL_MINOR_VERSION_INIT_OUT_SIZE {
@@ -811,26 +805,65 @@ impl<F: FileSystem + Sync, D: AsyncDrive> Server<F, D> {
     }
 
     pub(super) fn getlk<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.getlk() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let LkIn {
+            fh,
+            owner,
+            lk,
+            lk_flags,
+            ..
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+        match self.fs.getlk(
+            ctx.context(),
+            ctx.nodeid(),
+            fh.into(),
+            owner,
+            lk.into(),
+            lk_flags,
+        ) {
+            Ok(l) => ctx.reply_ok(Some(LkOut { lk: l.into() }), None),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
     pub(super) fn setlk<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.setlk() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let LkIn {
+            fh,
+            owner,
+            lk,
+            lk_flags,
+            ..
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+        match self.fs.setlk(
+            ctx.context(),
+            ctx.nodeid(),
+            fh.into(),
+            owner,
+            lk.into(),
+            lk_flags,
+        ) {
+            Ok(()) => ctx.reply_ok(None::<u8>, None),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
     pub(super) fn setlkw<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.setlkw() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let LkIn {
+            fh,
+            owner,
+            lk,
+            lk_flags,
+            ..
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+        match self.fs.setlk(
+            ctx.context(),
+            ctx.nodeid(),
+            fh.into(),
+            owner,
+            lk.into(),
+            lk_flags,
+        ) {
+            Ok(()) => ctx.reply_ok(None::<u8>, None),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
@@ -875,10 +908,13 @@ impl<F: FileSystem + Sync, D: AsyncDrive> Server<F, D> {
     pub(super) fn interrupt<S: BitmapSlice>(&self, _ctx: SrvContext<'_, F, D, S>) {}
 
     pub(super) fn bmap<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.bmap() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let BmapIn {
+            block, blocksize, ..
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+
+        match self.fs.bmap(ctx.context(), ctx.nodeid(), block, blocksize) {
+            Ok(block) => ctx.reply_ok(Some(BmapOut { block }), None),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
@@ -888,18 +924,74 @@ impl<F: FileSystem + Sync, D: AsyncDrive> Server<F, D> {
     }
 
     pub(super) fn ioctl<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.ioctl() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let IoctlIn {
+            fh,
+            flags,
+            cmd,
+            arg: _,
+            in_size,
+            out_size,
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+        // TODO: check fs capability of FUSE_CAP_IOCTL_DIR and return ENOTTY if unsupported.
+        let mut buf = IoctlData {
+            ..Default::default()
+        };
+        let in_size = in_size as usize;
+        // Make sure we have enough bytes to read the ioctl in buffer.
+        if in_size > ctx.r.available_bytes() {
+            return ctx.reply_error(io::Error::from_raw_os_error(libc::ENOTTY));
+        }
+        let mut data = vec![0u8; in_size];
+        if in_size > 0 {
+            let size = ctx.r.read(&mut data).map_err(Error::DecodeMessage)?;
+            if size > 0 {
+                buf.data = Some(&data[..size]);
+            }
+        }
+        match self.fs.ioctl(
+            ctx.context(),
+            ctx.nodeid(),
+            fh.into(),
+            flags,
+            cmd,
+            buf,
+            out_size,
+        ) {
+            Ok(res) => ctx.reply_ok(
+                Some(IoctlOut {
+                    result: res.result,
+                    ..Default::default()
+                }),
+                res.data,
+            ),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
     pub(super) fn poll<S: BitmapSlice>(&self, mut ctx: SrvContext<'_, F, D, S>) -> Result<usize> {
-        if let Err(e) = self.fs.poll() {
-            ctx.reply_error(e)
-        } else {
-            Ok(0)
+        let PollIn {
+            fh,
+            kh,
+            flags,
+            events,
+        } = ctx.r.read_obj().map_err(Error::DecodeMessage)?;
+
+        match self.fs.poll(
+            ctx.context(),
+            ctx.nodeid(),
+            fh.into(),
+            kh.into(),
+            flags,
+            events,
+        ) {
+            Ok(revents) => ctx.reply_ok(
+                Some(PollOut {
+                    revents,
+                    padding: 0,
+                }),
+                None,
+            ),
+            Err(e) => ctx.reply_error(e),
         }
     }
 
